unit HtmlUnit;

interface

uses
    SysUtils, Classes, Contnrs;

type
    //--------------------------------------------------------------------------
    //  \NX
    //--------------------------------------------------------------------------
    THighlight = class
    private
        FText: string;
        FFoundPos: integer;
    protected
        FOpenTag: string;
        FCloseTag: string;
        function    GetLength(): integer; virtual; abstract;
        function    Match(const S: string; const Offset: integer;
                          var FoundPos: integer): Boolean; virtual; abstract;
    public
        function    GetTaggedText(const S: string): string; virtual;
        constructor Create(const S, OpenTag, CloseTag: string); virtual;
        property Len: integer read GetLength;
        property FoundPos: integer read FFoundPos;
    end;

    THighlightClass = class of THighlight;

    //--------------------------------------------------------------------------
    //  sRgNX
    //--------------------------------------------------------------------------
    THighlightMultiComment = class(THighlight)
    private
        FMaeText: string;
        FAtoText: string;
        FLen: integer;
        FFirstMae: Boolean;
        FAtoExists: Boolean;
        FInLine: Boolean;
    protected
        function    GetLength(): integer; override;
        function    Match(const S: string; const Offset: integer;
                          var FoundPos: integer): Boolean; override;
    public
        constructor Create(const S, OpenTag, CloseTag: string); override;
        function    GetTaggedText(const S: string): string; override;
    end;

    //--------------------------------------------------------------------------
    //  萔NX
    //--------------------------------------------------------------------------
    THighlightString = class(THighlight)
    private
        FText: char;
        FLen: integer;
    protected
        function    GetLength(): integer; override;
        function    Match(const S: string; const Offset: integer;
                          var FoundPos: integer): Boolean; override;
    public
        constructor Create(const S, OpenTag, CloseTag: string); override;
    end;

    //--------------------------------------------------------------------------
    //  1sRgNX
    //--------------------------------------------------------------------------
    THighlightComment = class(THighlight)
    private
        FLen: integer;
    protected
        function    GetLength(): integer; override;
        function    Match(const S: string; const Offset: integer;
                          var FoundPos: integer): Boolean; override;
    end;

    //--------------------------------------------------------------------------
    //  \ꏈNX
    //--------------------------------------------------------------------------
    THighlightReserved = class(THighlight)
    protected
        function    GetLength(): integer; override;
        function    Match(const S: string; const Offset: integer;
                          var FoundPos: integer): Boolean; override;
    end;

    //--------------------------------------------------------------------------
    //  e
    //--------------------------------------------------------------------------
    TSqlToHtml = class
    private
        FWorkList: TStringList;
        FHighlightList: TObjectList;
        function    ConvertLine(const Line: string): string;
        procedure   LoadHighlightList();
    public
        constructor Create();
        destructor  Destroy(); override;
        procedure   Convert(SQL: TStrings; HTML: TStrings);
    end;

var
    SqlToHtml: TSqlToHtml = nil;

implementation

//-----------------------------------------------------------------------------
//  "<"">""&"AHTMLŐ\邽߂̕ϊ
function ChangeHTMLSpecialChar(text: string): string;
var
    i: integer;
    src, dest: array[0..2] of string;
begin

    //ȂƂȂ[vł܂킵Ȃ̂ŔzɁB
    //"&"ŏɕϊƂȂƑςȂƂ
    src[0] := '&';  dest[0] := '&amp;';
    src[1] := '<';  dest[1] := '&lt;';
    src[2] := '>';  dest[2] := '&gt;';

    //ϊRs[
    Result := Text;
    //bAfor i := 0 to 2 do
    for i := Low(src) to High(src) do
    begin
        //uĂ邾
        Result := StringReplace(Result, src[i], dest[i], [rfReplaceAll]);
    end;

end;

//-----------------------------------------------------------------------------
//  AvP[Vf[^tH_
function    GetDataDir(): string;
begin
    Result := ExtractFilePath(ParamStr(0)) + 'dat\';
end;

//*****************************************************************************
//  THighlight

//-----------------------------------------------------------------------------
//  RXgN^
constructor THighlight.Create(const S, OpenTag, CloseTag: string);
begin
    FText := S;
    FOpenTag := OpenTag;
    FCloseTag := CloseTag;
end;

//-----------------------------------------------------------------------------
//  ^OteLXg߂
function    THighlight.GetTaggedText(const S: string): string;
begin
    Result := FOpenTag + S + FCloseTag;
end;

//*****************************************************************************
//  THighlightMultiComment

//-----------------------------------------------------------------------------
//  RXgN^
constructor THighlightMultiComment.Create(const S, OpenTag, CloseTag: string);
begin
    //  S  /*,*/ ̂悤Ȍ`ł邱
    FMaeText := Copy(S, 1, Pos(',', S) - 1);
    FAtoText := Copy(S, Pos(',', S) + 1, Length(S));
    FOpenTag := OpenTag;
    FCloseTag := CloseTag;
    FFirstMae := false;
    FAtoExists:= false;
    FInLine   := false;
end;

//-----------------------------------------------------------------------------
//  ̒
function    THighlightMultiComment.GetLength(): integer;
begin
    Result := FLen;
end;

//-----------------------------------------------------------------------------
//  ^OteLXg߂
function    THighlightMultiComment.GetTaggedText(const S: string): string;
begin
    //  ^OȂꍇ
    Result := S;

    //  ߂ĊJneLXg𔭌
    if (FFirstMae) then
    begin
        Result := FOpenTag + Result;
        FFirstMae := false;
    end;

    //  IeLXg𔭌
    if (FAtoExists) then
    begin
        Result := Result + FCloseTag;
        FAtoExists := false;
    end;
end;

//-----------------------------------------------------------------------------
//  S̕񌟍
function    THighlightMultiComment.Match(const S: string;   //  ׂ1s
                  const Offset: integer;        //  Jnʒu
                  var FoundPos: integer         //  ʒu
                  ): Boolean;                   //  ?
var
    Target: string;
    Index, RealIndex: integer;
begin
    Result := false;

    //  ۂ̌Ώە
    Target := Copy(S, Offset, Length(S));

    //  ɕsRg[hɃWbNCĂ?
    if (FInLine) then
    begin
        //  Ƃ肠IeLXgKɒT
        Index := AnsiPos(UpperCase(FAtoText), UpperCase(Target));
        //  łȂI
        if (Index = 0) then
        begin
            //  ǂARg̒Ȃ̂łȂ
            FoundPos := 1;
            FFoundPos := FoundPos;
            FLen := Length(S);
            Result := true;
        end
        //  ܂
        else
        begin
            FoundPos := 1;
            FFoundPos := FoundPos;
            FLen := Index + Length(FAtoText) - 1;
            FAtoExists := true;
            FInLine := false;
            Result := true;
        end;
    end
    //  [
    else
    begin
        //  Ƃ肠JneLXgKɒT
        Index := AnsiPos(UpperCase(FMaeText), UpperCase(Target));
        //  łȂI
        if (Index > 0) then
        begin
            //  ۂ̈ʒu
            RealIndex := Offset + Index - 1;

            FFirstMae := true;
            FoundPos := RealIndex;
            FFoundPos := FoundPos;

            //  Ă
            Target := Copy(S, RealIndex + Length(FMaeText), Length(S));
            //  Ƃ肠IeLXgKɒT
            Index := AnsiPos(UpperCase(FAtoText), UpperCase(Target));
            //  Ȃł
            if (Index = 0) then
            begin
                //  s[h˓
                FInLine := true;
                FLen := Length(S) - FFoundPos + 1;
                Result := true;
            end
            //  ܂
            else
            begin
                //  ۂ̈ʒu
                RealIndex := RealIndex + Length(FMaeText) + Index - 1;

                FAtoExists := true;
                FLen := (RealIndex + Length(FAtoText) - 1) - FFoundPos + 1;
                Result := true;
            end;
        end
    end;
end;

//*****************************************************************************
//  THighlightString

//-----------------------------------------------------------------------------
//  RXgN^
constructor THighlightString.Create(const S, OpenTag, CloseTag: string);
begin
    FText := S[1];
    FOpenTag := OpenTag;
    FCloseTag := CloseTag;
end;

//-----------------------------------------------------------------------------
//  ̒
function    THighlightString.GetLength(): integer;
begin
    Result := FLen;
end;

//-----------------------------------------------------------------------------
//  S̕񌟍
function    THighlightString.Match(const S: string;   //  ׂ1s
                  const Offset: integer;        //  Jnʒu
                  var FoundPos: integer         //  ʒu
                  ): Boolean;                   //  ?
var
    Target: string;
    i, Index, RealIndex, LastPos: integer;
    c, nextc: char;
begin
    Result := false;

    //  ۂ̌Ώە
    Target := Copy(S, Offset, Length(S));

    //  Ƃ肠KɒT
    Index := AnsiPos(UpperCase(FText), UpperCase(Target));
    //  łȂI
    if (Index > 0) then
    begin
        //  ۂ̈ʒu
        RealIndex := Offset + Index - 1;
        //  Pcʒu
        LastPos := RealIndex; 
        //  Ă܂傤
        i := RealIndex + 1;
        while (i <= Length(S)) do
        begin
            //  1oCgłȂƑʖ
            if (ByteType(S, i) <> mbSingleByte) then
            begin
                Inc(i);
                continue;
            end;

            c := S[i];

            //  ̏I肾낤?
            if (c <> FText) then
            begin
                Inc(i);
                continue;
            end;

            //  2AƂĂȂ낤˂
            if (i < Length(S)) then
            begin
                nextc := S[i + 1];
                if (nextc = FText) then
                begin
                    Inc(i, 2);
                    continue;
                end;
            end;

            //  ܂ł݂
            LastPos := i;
            break;
        end;

        //  ܂ȂƂ?
        if (RealIndex = LastPos) then LastPos := Length(S);

        //  ǂ
        FLen := LastPos - RealIndex + 1;

        //@̑
        FoundPos := RealIndex;
        FFoundPos := FoundPos;
        Result := true;
    end;
end;

//*****************************************************************************
//  THighlightComment

//-----------------------------------------------------------------------------
//  ̒
function    THighlightComment.GetLength(): integer;
begin
    Result := FLen;
end;

//-----------------------------------------------------------------------------
//  S̕񌟍
function    THighlightComment.Match(const S: string;   //  ׂ1s
                  const Offset: integer;        //  Jnʒu
                  var FoundPos: integer         //  ʒu
                  ): Boolean;                   //  ?
var
    Target: string;
    Index, RealIndex, LastPos: integer;
begin
    Result := false;

    //  ۂ̌Ώە
    Target := Copy(S, Offset, Length(S));

    //  T
    Index := AnsiPos(UpperCase(FText), UpperCase(Target));
    //  łȂI
    if (Index > 0) then
    begin
        //  ۂ̈ʒu
        RealIndex := Offset + Index - 1;
        //  Ō܂łׂđΏ
        LastPos := Length(S); 
        //  ǂ
        FLen := LastPos - RealIndex + 1;
        //@̑
        FoundPos := RealIndex;
        FFoundPos := FoundPos;
        Result := true;
    end;
end;

//*****************************************************************************
//  THighlightReserved

//-----------------------------------------------------------------------------
//  ̒
function    THighlightReserved.GetLength(): integer;
begin
    Result := Length(FText);
end;

//-----------------------------------------------------------------------------
//  S̕񌟍
function    THighlightReserved.Match(const S: string;   //  ׂ1s
                  const Offset: integer;        //  Jnʒu
                  var FoundPos: integer         //  ʒu
                  ): Boolean;                   //  ?
var
    Target: string;
    Index, RealIndex: integer;
begin
    Result := false;
    //  ۂ̌Ώە
    Target := Copy(S, Offset, Length(S));

    while (Target <> '') do
    begin
        //  Ƃ肠KɒT
        Index := AnsiPos(UpperCase(FText), UpperCase(Target));
        //  łȂI
        if (Index = 0) then
        begin
            break;
        end
        //  ܂
        else
        begin
            //  ۂ̈ʒu
            RealIndex := Offset + Index - 1;
            //  O̕ɕςȂ̂˂?
            if (RealIndex > 1) then
            begin
                //  ȊO̕ʖڂł
                if not (S[RealIndex - 1] in [' ',#9,'(',')',';']) then
                begin
                    //  ͈͂ɂ炷
                    Target := Copy(Target, Index + Len, Length(S));
                    continue;
                end;
            end;
            //  ̕ɕςȂ̂˂?
            if (RealIndex + Len - 1 < Length(S)) then
            begin
                //  ȊO̕ʖڂł
                if not (S[RealIndex + Len] in [' ', #9, '(', ')',';']) then
                begin
                    //  ͈͂ɂ炷
                    Target := Copy(Target, Index + Len, Length(S));
                    continue;
                end;
            end;
            //  ̗p
            FoundPos := RealIndex;
            FFoundPos := FoundPos;
            Result := true;
            break;
        end;
    end;
end;

//*****************************************************************************
//  TSqlToHtml

//-----------------------------------------------------------------------------
//  RXgN^
constructor TSqlToHtml.Create();
begin
    //  Ɨp
    FWorkList := TStringList.Create();
    //  \ꏈNXXg
    FHighlightList := TObjectList.Create(true);
    //  \NX쐬
    LoadHighlightList();
end;
//-----------------------------------------------------------------------------
//  fXgN^
destructor  TSqlToHtml.Destroy();
begin
    FHighlightList.Free();
    FWorkList.Free();
end;
//-----------------------------------------------------------------------------
//  \ꗗǍ
procedure   TSqlToHtml.LoadHighlightList();

    //  ------------------------------
    function GetHighLightClass(const Name: string): THighlightClass;
    begin
        if (Name = 'MULTICOMMENT') then Result := THighlightMultiComment else
        if (Name = 'LINECOMMENT')  then Result := THighlightComment      else
        if (Name = 'STRING')       then Result := THighlightString       else
        if (Name = 'WORD')         then Result := THighlightReserved     else
        Result := nil;
    end;

    //  ------------------------------
    procedure CreateHighlights(HighlightClass: THighlightClass; const Path: string);
    var
        F: TextFile;
        S: string;
        OpenTag: string;
        CloseTag: string;
    begin
        AssignFile(F, Path);
        try
            //  J
            Reset(F);
            //  I肶Ȃ
            if (Eof(F)) then exit;

            //  1sځJn^O
            Readln(F, OpenTag);
            //  I肶Ȃ
            if (Eof(F)) then exit;

            //  2sځI^O
            Readln(F, CloseTag);

            //  cǍ
            while not Eof(F) do
            begin
                //  ǂ
                Readln(F, S);
                //  ȒPɃ`FbN
                if (S = '') then continue;
                if (HighlightClass = THighlightString) then
                    if (Length(S) <> 1) then continue;
                if (HighlightClass = THighlightMultiComment) then
                    if (Pos(',', S) = 0) then continue;
                //  NX쐬Ă
                FHighlightList.Add(HighlightClass.Create(S, OpenTag, CloseTag));
            end;
        finally
            CloseFile(F);
        end;
    end;

var
    Entry: TStringList;
    Helper: TStringList;
    HighlightClass: THighlightClass;
    EntryFile: string;

    i: integer;
begin

    //  ------------------------------
    //  Ô߃NA
    FHighlightList.Clear();

    //  ------------------------------
    //  ƗpStringList쐬
    Entry := TStringList.Create();
    Helper := TStringList.Create();

    //  ------------------------------
    //  Gg[t@CǍ
    Entry.LoadFromFile(GetDataDir() + '__entry.txt');

    //  ------------------------------
    //  es
    for i := 0 to Entry.Count - 1 do
    begin
        //  1s𕪉
        Helper.CommaText := Entry[i];

        //  \NXʃeLXgEt@C2łȂΔ΂
        if (Helper.Count <> 2) then continue;

        //  ΉNXȂĂ΂
        HighlightClass := GetHighLightClass(Helper[0]);
        if (HighlightClass = nil) then continue;

        //  t@C݂ĂȂĂ΂
        EntryFile := GetDataDir() + Helper[1];
        if (not FileExists(EntryFile)) then continue;

        //  Ă
        CreateHighlights(HighlightClass, EntryFile);
    end;

    //  ------------------------------
    //  
    Helper.Free;
    Entry.Free;
end;
//-----------------------------------------------------------------------------
//  ϊ
procedure   TSqlToHtml.Convert(SQL: TStrings; HTML: TStrings);
var
    i: integer;
begin
    HTML.Clear();
    for i := 0 to SQL.Count - 1 do
    begin
        HTML.Add(ConvertLine(SQL[i]));
    end;

    //  擪Ɩ<pre>ŋ
    if (HTML.Count > 0) then
    begin
        HTML.Insert(0, '<pre class="code">');
        HTML.Append('</pre>');
    end;
end;
//-----------------------------------------------------------------------------
//  1sϊ
function    TSqlToHtml.ConvertLine(const Line: string): string;
var
    i: integer;
    H, OkH: THighlight;
    Offset, FoundPos: integer;
    HiLiteText: string;
    aLine: string;
begin
    //  ꕶ͒uĂ(ȂƂłĂ..?)
    aLine := ChangeHTMLSpecialChar(Line);

    Result := '';

    //  Jnʒu
    Offset := 1;
    //  ̗pꂽ\
    OkH := nil;

    //  Jnʒu̒𒴂瓖RI
    while (Offset <= Length(aLine)) do
    begin

        //  e\Ɍ
        for i := 0 to FHighlightList.Count - 1 do
        begin
            H := THighlight(FHighlightList[i]);
            //  
            if (H.Match(aLine, Offset, FoundPos)) then
            begin
                //  O߂?
                if (OkH = nil) then
                begin
                    //  oĂ
                    OkH := H;
                end
                //  ߂ĂȂ
                else
                begin
                    //  O̓z葁ʒuŌ?
                    if (FoundPos < OkH.FoundPos) then
                    begin
                        //  O̗p
                        OkH := H;
                    end;
                end;
            end;
        end;
        //  ̂
        if (OkH <> nil) then
        begin
            //  \Ă݂悤
            HiLiteText := Copy(aLine, OkH.FoundPos, OkH.Len);
            HiLiteText := OkH.GetTaggedText(HiLiteText);

            //  ?
            Result := Result
            //  Jnʒu甭ʒu܂ł͖{
                    + Copy(aLine, Offset, OkH.FoundPos - Offset)
            //  ʒu甭܂ł͋\
                    + HiLiteText;
            //  ̗\͔\ŒT
            Offset := OkH.FoundPos + OkH.Len;
            OkH := nil;
        end
        //  Ȃ
        else
        begin
            Result := Result
                    + Copy(aLine, Offset, Length(aLine));
            //  ̍s͒אs
            break;
        end;
    end;
end;

//-----------------------------------------------------------------------------
initialization
    SqlToHtml := TSqlToHtml.Create();
//-----------------------------------------------------------------------------
finalization
    SqlToHtml.Free();
//-----------------------------------------------------------------------------
end.

